@@ -14,19 +14,22 @@ from mypy import build
 from mypy import defaults
 from mypy import state
 from mypy import util
-from mypy.modulefinder import BuildSource, FindModuleCache, mypy_path, SearchPaths
+from mypy.modulefinder import (
+    BuildSource, FindModuleCache, SearchPaths,
+    get_site_packages_dirs, mypy_path,
+)
 from mypy.find_sources import create_source_list, InvalidSourceList
 from mypy.fscache import FileSystemCache
 from mypy.errors import CompileError
 from mypy.errorcodes import error_codes
 from mypy.options import Options, BuildType
-from mypy.config_parser import parse_version, parse_config_file
+from mypy.config_parser import get_config_module_names, parse_version, parse_config_file
 from mypy.split_namespace import SplitNamespace
 
 from mypy.version import __version__
 
-orig_stat = os.stat  # type: Final
-MEM_PROFILE = False  # type: Final  # If True, dump memory profile
+orig_stat: Final = os.stat
+MEM_PROFILE: Final = False  # If True, dump memory profile
 
 
 def stat_proxy(path: str) -> os.stat_result:
@@ -45,13 +48,16 @@ def main(script_path: Optional[str],
          stdout: TextIO,
          stderr: TextIO,
          args: Optional[List[str]] = None,
+         clean_exit: bool = False,
          ) -> None:
     """Main entry point to the type checker.
 
     Args:
         script_path: Path to the 'mypy' script (used for finding data files).
         args: Custom command-line arguments.  If not given, sys.argv[1:] will
-        be used.
+            be used.
+        clean_exit: Don't hard kill the process on exit. This allows catching
+            SystemExit.
     """
     util.check_python_version('mypy')
     t0 = time.time()
@@ -63,42 +69,41 @@ def main(script_path: Optional[str],
     fscache = FileSystemCache()
     sources, options = process_options(args, stdout=stdout, stderr=stderr,
                                        fscache=fscache)
+    if clean_exit:
+        options.fast_exit = False
 
-    messages = []
     formatter = util.FancyFormatter(stdout, stderr, options.show_error_codes)
 
-    def flush_errors(new_messages: List[str], serious: bool) -> None:
-        if options.pretty:
-            new_messages = formatter.fit_in_terminal(new_messages)
-        messages.extend(new_messages)
-        f = stderr if serious else stdout
-        try:
-            for msg in new_messages:
-                if options.color_output:
-                    msg = formatter.colorize(msg)
-                f.write(msg + '\n')
-            f.flush()
-        except BrokenPipeError:
-            sys.exit(2)
+    if options.install_types and (stdout is not sys.stdout or stderr is not sys.stderr):
+        # Since --install-types performs user input, we want regular stdout and stderr.
+        fail("error: --install-types not supported in this mode of running mypy", stderr, options)
 
-    serious = False
-    blockers = False
-    res = None
-    try:
-        # Keep a dummy reference (res) for memory profiling below, as otherwise
-        # the result could be freed.
-        res = build.build(sources, options, None, flush_errors, fscache, stdout, stderr)
-    except CompileError as e:
-        blockers = True
-        if not e.use_stdout:
-            serious = True
-    if options.warn_unused_configs and options.unused_configs and not options.incremental:
-        print("Warning: unused section(s) in %s: %s" %
-              (options.config_file,
-               ", ".join("[mypy-%s]" % glob for glob in options.per_module_options.keys()
-                         if glob in options.unused_configs)),
-              file=stderr)
-    maybe_write_junit_xml(time.time() - t0, serious, messages, options)
+    if options.non_interactive and not options.install_types:
+        fail("error: --non-interactive is only supported with --install-types", stderr, options)
+
+    if options.install_types and not options.incremental:
+        fail("error: --install-types not supported with incremental mode disabled",
+             stderr, options)
+
+    if options.install_types and options.python_executable is None:
+        fail("error: --install-types not supported without python executable or site packages",
+             stderr, options)
+
+    if options.install_types and not sources:
+        install_types(formatter, options, non_interactive=options.non_interactive)
+        return
+
+    res, messages, blockers = run_build(sources, options, fscache, t0, stdout, stderr)
+
+    if options.non_interactive:
+        missing_pkgs = read_types_packages_to_install(options.cache_dir, after_run=True)
+        if missing_pkgs:
+            # Install missing type packages and rerun build.
+            install_types(formatter, options, after_run=True, non_interactive=True)
+            fscache.flush()
+            print()
+            res, messages, blockers = run_build(sources, options, fscache, t0, stdout, stderr)
+        show_messages(messages, stderr, formatter, options)
 
     if MEM_PROFILE:
         from mypy.memprofile import print_memory_profile
@@ -108,15 +113,25 @@ def main(script_path: Optional[str],
     if messages:
         code = 2 if blockers else 1
     if options.error_summary:
-        if messages:
-            n_errors, n_files = util.count_stats(messages)
-            if n_errors:
-                stdout.write(formatter.format_error(n_errors, n_files, len(sources),
-                                                    options.color_output) + '\n')
-        else:
-            stdout.write(formatter.format_success(len(sources),
-                                                  options.color_output) + '\n')
+        n_errors, n_notes, n_files = util.count_stats(messages)
+        if n_errors:
+            summary = formatter.format_error(
+                n_errors, n_files, len(sources), blockers=blockers,
+                use_color=options.color_output
+            )
+            stdout.write(summary + '\n')
+        # Only notes should also output success
+        elif not messages or n_notes == len(messages):
+            stdout.write(formatter.format_success(len(sources), options.color_output) + '\n')
         stdout.flush()
+
+    if options.install_types and not options.non_interactive:
+        result = install_types(formatter, options, after_run=True, non_interactive=False)
+        if result:
+            print()
+            print("note: Run mypy again for up-to-date results with installed types")
+            code = 2
+
     if options.fast_exit:
         # Exit without freeing objects -- it's faster.
         #
@@ -129,29 +144,83 @@ def main(script_path: Optional[str],
     list([res])
 
 
+def run_build(sources: List[BuildSource],
+              options: Options,
+              fscache: FileSystemCache,
+              t0: float,
+              stdout: TextIO,
+              stderr: TextIO) -> Tuple[Optional[build.BuildResult], List[str], bool]:
+    formatter = util.FancyFormatter(stdout, stderr, options.show_error_codes)
+
+    messages = []
+
+    def flush_errors(new_messages: List[str], serious: bool) -> None:
+        if options.pretty:
+            new_messages = formatter.fit_in_terminal(new_messages)
+        messages.extend(new_messages)
+        if options.non_interactive:
+            # Collect messages and possibly show them later.
+            return
+        f = stderr if serious else stdout
+        show_messages(new_messages, f, formatter, options)
+
+    serious = False
+    blockers = False
+    res = None
+    try:
+        # Keep a dummy reference (res) for memory profiling afterwards, as otherwise
+        # the result could be freed.
+        res = build.build(sources, options, None, flush_errors, fscache, stdout, stderr)
+    except CompileError as e:
+        blockers = True
+        if not e.use_stdout:
+            serious = True
+    if (options.warn_unused_configs
+            and options.unused_configs
+            and not options.incremental
+            and not options.non_interactive):
+        print("Warning: unused section(s) in %s: %s" %
+              (options.config_file,
+              get_config_module_names(options.config_file,
+                                      [glob for glob in options.per_module_options.keys()
+                                      if glob in options.unused_configs])),
+              file=stderr)
+    maybe_write_junit_xml(time.time() - t0, serious, messages, options)
+    return res, messages, blockers
+
+
+def show_messages(messages: List[str],
+                  f: TextIO,
+                  formatter: util.FancyFormatter,
+                  options: Options) -> None:
+    for msg in messages:
+        if options.color_output:
+            msg = formatter.colorize(msg)
+        f.write(msg + '\n')
+    f.flush()
+
+
 # Make the help output a little less jarring.
 class AugmentedHelpFormatter(argparse.RawDescriptionHelpFormatter):
     def __init__(self, prog: str) -> None:
         super().__init__(prog=prog, max_help_position=28)
 
-    # FIXME: typeshed incorrectly has the type of indent as int when
-    # it should be str. Make it Any to avoid rusing mypyc.
-    def _fill_text(self, text: str, width: int, indent: Any) -> str:
+    def _fill_text(self, text: str, width: int, indent: str) -> str:
         if '\n' in text:
             # Assume we want to manually format the text
             return super()._fill_text(text, width, indent)
         else:
-            # Assume we want argparse to manage wrapping, indentating, and
+            # Assume we want argparse to manage wrapping, indenting, and
             # formatting the text for us.
             return argparse.HelpFormatter._fill_text(self, text, width, indent)
 
 
 # Define pairs of flag prefixes with inverse meaning.
-flag_prefix_pairs = [
+flag_prefix_pairs: Final = [
     ('allow', 'disallow'),
     ('show', 'hide'),
-]  # type: Final
-flag_prefix_map = {}  # type: Final[Dict[str, str]]
+]
+flag_prefix_map: Final[Dict[str, str]] = {}
 for a, b in flag_prefix_pairs:
     flag_prefix_map[a] = b
     flag_prefix_map[b] = a
@@ -176,8 +245,8 @@ class PythonExecutableInferenceError(Exception):
 def python_executable_prefix(v: str) -> List[str]:
     if sys.platform == 'win32':
         # on Windows, all Python executables are named `python`. To handle this, there
-        # is the `py` launcher, which can be passed a version e.g. `py -3.5`, and it will
-        # execute an installed Python 3.5 interpreter. See also:
+        # is the `py` launcher, which can be passed a version e.g. `py -3.8`, and it will
+        # execute an installed Python 3.8 interpreter. See also:
         # https://docs.python.org/3/using/windows.html#python-launcher-for-windows
         return ['py', '-{}'.format(v)]
     else:
@@ -221,11 +290,11 @@ def infer_python_executable(options: Options,
     options.python_executable = python_executable
 
 
-HEADER = """%(prog)s [-h] [-v] [-V] [more options; see below]
-            [-m MODULE] [-p PACKAGE] [-c PROGRAM_TEXT] [files ...]"""  # type: Final
+HEADER: Final = """%(prog)s [-h] [-v] [-V] [more options; see below]
+            [-m MODULE] [-p PACKAGE] [-c PROGRAM_TEXT] [files ...]"""
 
 
-DESCRIPTION = """
+DESCRIPTION: Final = """
 Mypy is a program that will type check your Python code.
 
 Pass in any files or folders you want to type check. Mypy will
@@ -235,22 +304,22 @@ recursively traverse any provided folders to find .py files:
 
 For more information on getting started, see:
 
-- http://mypy.readthedocs.io/en/latest/getting_started.html
+- https://mypy.readthedocs.io/en/stable/getting_started.html
 
 For more details on both running mypy and using the flags below, see:
 
-- http://mypy.readthedocs.io/en/latest/running_mypy.html
-- http://mypy.readthedocs.io/en/latest/command_line.html
+- https://mypy.readthedocs.io/en/stable/running_mypy.html
+- https://mypy.readthedocs.io/en/stable/command_line.html
 
 You can also use a config file to configure mypy instead of using
 command line flags. For more details, see:
 
-- http://mypy.readthedocs.io/en/latest/config_file.html
-"""  # type: Final
+- https://mypy.readthedocs.io/en/stable/config_file.html
+"""
 
-FOOTER = """Environment variables:
+FOOTER: Final = """Environment variables:
   Define MYPYPATH for additional module search path entries.
-  Define MYPY_CACHE_DIR to override configuration cache_dir path."""  # type: Final
+  Define MYPY_CACHE_DIR to override configuration cache_dir path."""
 
 
 class CapturableArgumentParser(argparse.ArgumentParser):
@@ -373,8 +442,8 @@ def process_options(args: List[str],
                                       stdout=stdout,
                                       stderr=stderr)
 
-    strict_flag_names = []  # type: List[str]
-    strict_flag_assignments = []  # type: List[Tuple[str, bool]]
+    strict_flag_names: List[str] = []
+    strict_flag_assignments: List[Tuple[str, bool]] = []
 
     def add_invertible_flag(flag: str,
                             *,
@@ -423,9 +492,11 @@ def process_options(args: List[str],
     general_group.add_argument(
         '-v', '--verbose', action='count', dest='verbosity',
         help="More verbose messages")
+
+    compilation_status = "no" if __file__.endswith(".py") else "yes"
     general_group.add_argument(
         '-V', '--version', action=CapturableVersionAction,
-        version='%(prog)s ' + __version__,
+        version='%(prog)s ' + __version__ + f" (compiled: {compilation_status})",
         help="Show program's version number and exit",
         stdout=stdout)
 
@@ -439,7 +510,8 @@ def process_options(args: List[str],
         help="Configuration file, must have a [mypy] section "
              "(defaults to {})".format(', '.join(defaults.CONFIG_FILES)))
     add_invertible_flag('--warn-unused-configs', default=False, strict_flag=True,
-                        help="Warn about unused '[mypy-<pattern>]' config sections",
+                        help="Warn about unused '[mypy-<pattern>]' or '[[tool.mypy.overrides]]' "
+                             "config sections",
                         group=config_group)
 
     imports_group = parser.add_argument_group(
@@ -544,7 +616,7 @@ def process_options(args: List[str],
         title='None and Optional handling',
         description="Adjust how values of type 'None' are handled. For more context on "
                     "how mypy handles values of type 'None', see: "
-                    "http://mypy.readthedocs.io/en/latest/kinds_of_types.html#no-strict-optional")
+                    "https://mypy.readthedocs.io/en/stable/kinds_of_types.html#no-strict-optional")
     add_invertible_flag('--no-implicit-optional', default=False, strict_flag=True,
                         help="Don't assume arguments with default values of None are Optional",
                         group=none_group)
@@ -576,7 +648,7 @@ def process_options(args: List[str],
                         group=lint_group)
     add_invertible_flag('--warn-unreachable', default=False, strict_flag=False,
                         help="Warn about statements or expressions inferred to be"
-                             " unreachable or redundant",
+                             " unreachable",
                         group=lint_group)
 
     # Note: this group is intentionally added here even though we don't add
@@ -607,6 +679,10 @@ def process_options(args: List[str],
                              " non-overlapping types",
                         group=strictness_group)
 
+    add_invertible_flag('--strict-concatenate', default=False, strict_flag=True,
+                        help="Make arguments prepended via Concatenate be truly positional-only",
+                        group=strictness_group)
+
     strict_help = "Strict mode; enables the following flags: {}".format(
         ", ".join(strict_flag_names))
     strictness_group.add_argument(
@@ -648,6 +724,8 @@ def process_options(args: List[str],
     add_invertible_flag('--show-absolute-path', default=False,
                         help="Show absolute paths to files",
                         group=error_group)
+    error_group.add_argument('--soft-error-limit', default=defaults.MANY_ERRORS_THRESHOLD,
+                             type=int, dest="many_errors_threshold", help=argparse.SUPPRESS)
 
     incremental_group = parser.add_argument_group(
         title='Incremental mode',
@@ -655,7 +733,7 @@ def process_options(args: List[str],
                     "Mypy caches type information about modules into a cache to "
                     "let you speed up future invocations of mypy. Also see "
                     "mypy's daemon mode: "
-                    "mypy.readthedocs.io/en/latest/mypy_daemon.html#mypy-daemon")
+                    "mypy.readthedocs.io/en/stable/mypy_daemon.html#mypy-daemon")
     incremental_group.add_argument(
         '-i', '--incremental', action='store_true',
         help=argparse.SUPPRESS)
@@ -705,7 +783,7 @@ def process_options(args: List[str],
         dest='shadow_file', action='append',
         help="When encountering SOURCE_FILE, read and type check "
              "the contents of SHADOW_FILE instead.")
-    add_invertible_flag('--fast-exit', default=False, help=argparse.SUPPRESS,
+    add_invertible_flag('--fast-exit', default=True, help=argparse.SUPPRESS,
                         group=internals_group)
 
     report_group = parser.add_argument_group(
@@ -731,6 +809,14 @@ def process_options(args: List[str],
         '--scripts-are-modules', action='store_true',
         help="Script x becomes module x instead of __main__")
 
+    add_invertible_flag('--install-types', default=False, strict_flag=False,
+                        help="Install detected missing library stub packages using pip",
+                        group=other_group)
+    add_invertible_flag('--non-interactive', default=False, strict_flag=False,
+                        help=("Install stubs without asking for confirmation and hide " +
+                              "errors, with --install-types"),
+                        group=other_group, inverse="--interactive")
+
     if server_options:
         # TODO: This flag is superfluous; remove after a short transition (2018-03-16)
         other_group.add_argument(
@@ -780,12 +866,29 @@ def process_options(args: List[str],
     # Must be followed by another flag or by '--' (and then only file args may follow).
     parser.add_argument('--cache-map', nargs='+', dest='special-opts:cache_map',
                         help=argparse.SUPPRESS)
+    parser.add_argument('--enable-incomplete-features', default=False,
+                        help=argparse.SUPPRESS)
 
     # options specifying code to check
     code_group = parser.add_argument_group(
         title="Running code",
         description="Specify the code you want to type check. For more details, see "
-                    "mypy.readthedocs.io/en/latest/running_mypy.html#running-mypy")
+                    "mypy.readthedocs.io/en/stable/running_mypy.html#running-mypy")
+    add_invertible_flag(
+        '--explicit-package-bases', default=False,
+        help="Use current directory and MYPYPATH to determine module names of files passed",
+        group=code_group)
+    code_group.add_argument(
+        "--exclude",
+        action="append",
+        metavar="PATTERN",
+        default=[],
+        help=(
+            "Regular expression to match file names, directory names or paths which mypy should "
+            "ignore while recursively discovering files to check, e.g. --exclude '/setup\\.py$'. "
+            "May be specified more than once, eg. --exclude a --exclude b"
+        )
+    )
     code_group.add_argument(
         '-m', '--module', action='append', metavar='MODULE',
         default=[],
@@ -832,6 +935,7 @@ def process_options(args: List[str],
     environ_cache_dir = os.getenv('MYPY_CACHE_DIR', '')
     if environ_cache_dir.strip():
         options.cache_dir = environ_cache_dir
+    options.cache_dir = os.path.expanduser(options.cache_dir)
 
     # Parse command line for real, using a split namespace.
     special_opts = argparse.Namespace()
@@ -848,9 +952,9 @@ def process_options(args: List[str],
     if special_opts.no_executable or options.no_site_packages:
         options.python_executable = None
 
-    # Paths listed in the config file will be ignored if any paths are passed on
-    # the command line.
-    if options.files and not special_opts.files:
+    # Paths listed in the config file will be ignored if any paths, modules or packages
+    # are passed on the command line.
+    if options.files and not (special_opts.files or special_opts.packages or special_opts.modules):
         special_opts.files = options.files
 
     # Check for invalid argument combinations.
@@ -858,10 +962,15 @@ def process_options(args: List[str],
         code_methods = sum(bool(c) for c in [special_opts.modules + special_opts.packages,
                                              special_opts.command,
                                              special_opts.files])
-        if code_methods == 0:
+        if code_methods == 0 and not options.install_types:
             parser.error("Missing target module, package, files, or command.")
         elif code_methods > 1:
             parser.error("May only specify one of: module/package, files, or command.")
+    if options.explicit_package_bases and not options.namespace_packages:
+        parser.error(
+            "Can only use --explicit-package-bases with --namespace-packages, since otherwise "
+            "examining __init__.py's is sufficient to determine module names for files"
+        )
 
     # Check for overlapping `--always-true` and `--always-false` flags.
     overlap = set(options.always_true) & set(options.always_false)
@@ -916,6 +1025,11 @@ def process_options(args: List[str],
 
         process_cache_map(parser, special_opts, options)
 
+    # An explicitly specified cache_fine_grained implies local_partial_types
+    # (because otherwise the cache is not compatible with dmypy)
+    if options.cache_fine_grained:
+        options.local_partial_types = True
+
     # Let logical_deps imply cache_fine_grained (otherwise the former is useless).
     if options.logical_deps:
         options.cache_fine_grained = True
@@ -923,10 +1037,14 @@ def process_options(args: List[str],
     # Set target.
     if special_opts.modules + special_opts.packages:
         options.build_type = BuildType.MODULE
-        search_paths = SearchPaths((os.getcwd(),), tuple(mypy_path() + options.mypy_path), (), ())
+        egg_dirs, site_packages = get_site_packages_dirs(options.python_executable)
+        search_paths = SearchPaths((os.getcwd(),),
+                                   tuple(mypy_path() + options.mypy_path),
+                                   tuple(egg_dirs + site_packages),
+                                   ())
         targets = []
         # TODO: use the same cache that the BuildManager will
-        cache = FindModuleCache(search_paths, fscache, options, special_opts.packages)
+        cache = FindModuleCache(search_paths, fscache, options)
         for p in special_opts.packages:
             if os.sep in p or os.altsep and os.altsep in p:
                 fail("Package name '{}' cannot have a slash in it.".format(p),
@@ -976,12 +1094,12 @@ def process_package_roots(fscache: Optional[FileSystemCache],
         # Empty package root is always okay.
         if root:
             root = os.path.relpath(root)  # Normalize the heck out of it.
+            if not root.endswith(os.sep):
+                root = root + os.sep
             if root.startswith(dotdotslash):
                 parser.error("Package root cannot be above current directory: %r" % root)
             if root in trivial_paths:
                 root = ''
-            elif not root.endswith(os.sep):
-                root = root + os.sep
         package_root.append(root)
     options.package_root = package_root
     # Pass the package root on the the filesystem cache.
@@ -1022,3 +1140,50 @@ def fail(msg: str, stderr: TextIO, options: Options) -> None:
     stderr.write('%s\n' % msg)
     maybe_write_junit_xml(0.0, serious=True, messages=[msg], options=options)
     sys.exit(2)
+
+
+def read_types_packages_to_install(cache_dir: str, after_run: bool) -> List[str]:
+    if not os.path.isdir(cache_dir):
+        if not after_run:
+            sys.stderr.write(
+                "error: Can't determine which types to install with no files to check " +
+                "(and no cache from previous mypy run)\n"
+            )
+        else:
+            sys.stderr.write(
+                "error: --install-types failed (no mypy cache directory)\n"
+            )
+        sys.exit(2)
+    fnam = build.missing_stubs_file(cache_dir)
+    if not os.path.isfile(fnam):
+        # No missing stubs.
+        return []
+    with open(fnam) as f:
+        return [line.strip() for line in f.readlines()]
+
+
+def install_types(formatter: util.FancyFormatter,
+                  options: Options,
+                  *,
+                  after_run: bool = False,
+                  non_interactive: bool = False) -> bool:
+    """Install stub packages using pip if some missing stubs were detected."""
+    packages = read_types_packages_to_install(options.cache_dir, after_run)
+    if not packages:
+        # If there are no missing stubs, generate no output.
+        return False
+    if after_run and not non_interactive:
+        print()
+    print('Installing missing stub packages:')
+    assert options.python_executable, 'Python executable required to install types'
+    cmd = [options.python_executable, '-m', 'pip', 'install'] + packages
+    print(formatter.style(' '.join(cmd), 'none', bold=True))
+    print()
+    if not non_interactive:
+        x = input('Install? [yN] ')
+        if not x.strip() or not x.lower().startswith('y'):
+            print(formatter.style('mypy: Skipping installation', 'red', bold=True))
+            sys.exit(2)
+        print()
+    subprocess.run(cmd)
+    return True
