@@ -1,9 +1,9 @@
-"""Top-level logic for the new semantic analyzer.
+"""Top-level logic for the semantic analyzer.
 
 The semantic analyzer binds names, resolves imports, detects various
 special constructs that don't have dedicated AST nodes after parse
-(such as 'cast' which looks like a call), and performs various simple
-consistency checks.
+(such as 'cast' which looks like a call), populates symbol tables, and
+performs various simple consistency checks.
 
 Semantic analysis of each SCC (strongly connected component; import
 cycle) is performed in one unit. Each module is analyzed as multiple
@@ -24,15 +24,15 @@ deferral if they can't be satisfied. Initially every module in the SCC
 will be incomplete.
 """
 
-import contextlib
-from typing import List, Tuple, Optional, Union, Callable, Iterator
-from typing_extensions import TYPE_CHECKING
+from typing import List, Tuple, Optional, Union, Callable
+from typing_extensions import TYPE_CHECKING, Final, TypeAlias as _TypeAlias
 
+from mypy.backports import nullcontext
 from mypy.nodes import (
     MypyFile, TypeInfo, FuncDef, Decorator, OverloadedFuncDef, Var
 )
 from mypy.semanal_typeargs import TypeArgumentAnalyzer
-from mypy.state import strict_optional_set
+import mypy.state
 from mypy.semanal import (
     SemanticAnalyzer, apply_semantic_analyzer_patches, remove_imported_names_from_symtable
 )
@@ -44,22 +44,23 @@ from mypy.errors import Errors
 from mypy.semanal_infer import infer_decorator_signature_if_simple
 from mypy.checker import FineGrainedDeferredNode
 from mypy.server.aststrip import SavedAttributes
+from mypy.util import is_typeshed_file
 import mypy.build
 
 if TYPE_CHECKING:
     from mypy.build import Graph, State
 
 
-Patches = List[Tuple[int, Callable[[], None]]]
+Patches: _TypeAlias = List[Tuple[int, Callable[[], None]]]
 
 
 # If we perform this many iterations, raise an exception since we are likely stuck.
-MAX_ITERATIONS = 20
+MAX_ITERATIONS: Final = 20
 
 
 # Number of passes over core modules before going on to the rest of the builtin SCC.
-CORE_WARMUP = 2
-core_modules = ['typing', 'builtins', 'abc', 'collections']
+CORE_WARMUP: Final = 2
+core_modules: Final = ['typing', 'builtins', 'abc', 'collections']
 
 
 def semantic_analysis_for_scc(graph: 'Graph', scc: List[str], errors: Errors) -> None:
@@ -70,7 +71,7 @@ def semantic_analysis_for_scc(graph: 'Graph', scc: List[str], errors: Errors) ->
     The scc will be processed roughly in the order the modules are included
     in the list.
     """
-    patches = []  # type: Patches
+    patches: Patches = []
     # Note that functions can't define new module-level attributes
     # using 'global x', since module top levels are fully processed
     # before functions. This limitation is unlikely to go away soon.
@@ -115,7 +116,7 @@ def semantic_analysis_for_targets(
     defined on self) removed by AST stripper that may need to be reintroduced
     here.  They must be added before any methods are analyzed.
     """
-    patches = []  # type: Patches
+    patches: Patches = []
     if any(isinstance(n.node, MypyFile) for n in nodes):
         # Process module top level first (if needed).
         process_top_levels(graph, [state.id], patches)
@@ -126,7 +127,7 @@ def semantic_analysis_for_targets(
             # Already done above.
             continue
         process_top_level_function(analyzer, state, state.id,
-                                   n.node.fullname(), n.node, n.active_typeinfo, patches)
+                                   n.node.fullname, n.node, n.active_typeinfo, patches)
     apply_semantic_analyzer_patches(patches)
 
     check_type_arguments_in_targets(nodes, state, state.manager.errors)
@@ -189,7 +190,7 @@ def process_top_levels(graph: 'Graph', scc: List[str], patches: Patches) -> None
         if final_iteration:
             # Give up. It's impossible to bind all names.
             state.manager.incomplete_namespaces.clear()
-        all_deferred = []  # type: List[str]
+        all_deferred: List[str] = []
         any_progress = False
         while worklist:
             next_id = worklist.pop()
@@ -248,7 +249,7 @@ def process_top_level_function(analyzer: 'SemanticAnalyzer',
     """Analyze single top-level function or method.
 
     Process the body of the function (including nested functions) again and again,
-    until all names have been resolved (ot iteration limit reached).
+    until all names have been resolved (or iteration limit reached).
     """
     # We need one more iteration after incomplete is False (e.g. to report errors, if any).
     final_iteration = False
@@ -288,7 +289,7 @@ TargetInfo = Tuple[str, Union[MypyFile, FuncDef, OverloadedFuncDef, Decorator],
 
 def get_all_leaf_targets(file: MypyFile) -> List[TargetInfo]:
     """Return all leaf targets in a symbol table (module-level and methods)."""
-    result = []  # type: List[TargetInfo]
+    result: List[TargetInfo] = []
     for fullname, node, active_type in file.local_definitions():
         if isinstance(node.node, (FuncDef, OverloadedFuncDef, Decorator)):
             result.append((fullname, node.node, active_type))
@@ -305,8 +306,8 @@ def semantic_analyze_target(target: str,
 
     Return tuple with these items:
     - list of deferred targets
-    - was some definition incomplete
-    - were any new names were defined (or placeholders replaced)
+    - was some definition incomplete (need to run another pass)
+    - were any new names defined (or placeholders replaced)
     """
     state.manager.processed_targets.append(target)
     tree = state.tree
@@ -335,6 +336,12 @@ def semantic_analyze_target(target: str,
         priority = mypy.build.PRI_LOW
         if priority <= state.priorities.get(dep, priority):
             state.priorities[dep] = priority
+
+    # Clear out some stale data to avoid memory leaks and astmerge
+    # validity check confusion
+    analyzer.statement = None
+    del analyzer.cur_mod_node
+
     if analyzer.deferred:
         return [target], analyzer.incomplete, analyzer.progress
     else:
@@ -347,9 +354,9 @@ def check_type_arguments(graph: 'Graph', scc: List[str], errors: Errors) -> None
         assert state.tree
         analyzer = TypeArgumentAnalyzer(errors,
                                         state.options,
-                                        errors.is_typeshed_file(state.path or ''))
+                                        is_typeshed_file(state.path or ''))
         with state.wrap_context():
-            with strict_optional_set(state.options.strict_optional):
+            with mypy.state.state.strict_optional_set(state.options.strict_optional):
                 state.tree.accept(analyzer)
 
 
@@ -362,15 +369,15 @@ def check_type_arguments_in_targets(targets: List[FineGrainedDeferredNode], stat
     """
     analyzer = TypeArgumentAnalyzer(errors,
                                     state.options,
-                                    errors.is_typeshed_file(state.path or ''))
+                                    is_typeshed_file(state.path or ''))
     with state.wrap_context():
-        with strict_optional_set(state.options.strict_optional):
+        with mypy.state.state.strict_optional_set(state.options.strict_optional):
             for target in targets:
-                func = None  # type: Optional[Union[FuncDef, OverloadedFuncDef]]
+                func: Optional[Union[FuncDef, OverloadedFuncDef]] = None
                 if isinstance(target.node, (FuncDef, OverloadedFuncDef)):
                     func = target.node
                 saved = (state.id, target.active_typeinfo, func)  # module, class, function
-                with errors.scope.saved_scope(saved) if errors.scope else nothing():
+                with errors.scope.saved_scope(saved) if errors.scope else nullcontext():
                     analyzer.recurse_into_functions = func is not None
                     target.node.accept(analyzer)
 
@@ -382,7 +389,7 @@ def calculate_class_properties(graph: 'Graph', scc: List[str], errors: Errors) -
         for _, node, _ in tree.local_definitions():
             if isinstance(node.node, TypeInfo):
                 saved = (module, node.node, None)  # module, class, function
-                with errors.scope.saved_scope(saved) if errors.scope else nothing():
+                with errors.scope.saved_scope(saved) if errors.scope else nullcontext():
                     calculate_class_abstract_status(node.node, tree.is_stub, errors)
                     check_protocol_status(node.node, errors)
                     calculate_class_vars(node.node)
@@ -392,8 +399,3 @@ def calculate_class_properties(graph: 'Graph', scc: List[str], errors: Errors) -
 def check_blockers(graph: 'Graph', scc: List[str]) -> None:
     for module in scc:
         graph[module].check_blockers()
-
-
-@contextlib.contextmanager
-def nothing() -> Iterator[None]:
-    yield
