@@ -21,11 +21,12 @@ from typing_extensions import Type
 
 import mypy.build
 import mypy.modulefinder
+import mypy.state
 import mypy.types
 from mypy import nodes
 from mypy.config_parser import parse_config_file
 from mypy.options import Options
-from mypy.util import FancyFormatter
+from mypy.util import FancyFormatter, bytes_to_human_readable_repr, is_dunder
 
 
 class Missing:
@@ -144,6 +145,11 @@ class Error:
         return "".join(output)
 
 
+# ====================
+# Core logic
+# ====================
+
+
 def test_module(module_name: str) -> Iterator[Error]:
     """Tests a given module's stub against introspecting it at runtime.
 
@@ -204,12 +210,15 @@ def verify_mypyfile(
     to_check = set(
         m
         for m, o in stub.names.items()
-        if not o.module_hidden and (not m.startswith("_") or hasattr(runtime, m))
+        if not o.module_hidden and (not is_probably_private(m) or hasattr(runtime, m))
     )
 
     def _belongs_to_runtime(r: types.ModuleType, attr: str) -> bool:
         obj = getattr(r, attr)
-        obj_mod = getattr(obj, "__module__", None)
+        try:
+            obj_mod = getattr(obj, "__module__", None)
+        except Exception:
+            return False
         if obj_mod is not None:
             return obj_mod == r.__name__
         return not isinstance(obj, types.ModuleType)
@@ -220,7 +229,7 @@ def verify_mypyfile(
         else [
             m
             for m in dir(runtime)
-            if not m.startswith("_")
+            if not is_probably_private(m)
             # Ensure that the object's module is `runtime`, since in the absence of __all__ we
             # don't have a good way to detect re-exports at runtime.
             and _belongs_to_runtime(runtime, m)
@@ -228,7 +237,7 @@ def verify_mypyfile(
     )
     # Check all things declared in module's __all__, falling back to our best guess
     to_check.update(runtime_public_contents)
-    to_check.difference_update({"__file__", "__doc__", "__name__", "__builtins__", "__package__"})
+    to_check.difference_update(IGNORED_MODULE_DUNDERS)
 
     for entry in sorted(to_check):
         stub_entry = stub.names[entry].node if entry in stub.names else MISSING
@@ -236,11 +245,19 @@ def verify_mypyfile(
             # Don't recursively check exported modules, since that leads to infinite recursion
             continue
         assert stub_entry is not None
-        yield from verify(
-            stub_entry,
-            getattr(runtime, entry, MISSING),
-            object_path + [entry],
-        )
+        try:
+            runtime_entry = getattr(runtime, entry, MISSING)
+        except Exception:
+            # Catch all exceptions in case the runtime raises an unexpected exception
+            # from __getattr__ or similar.
+            continue
+        yield from verify(stub_entry, runtime_entry, object_path + [entry])
+
+
+if sys.version_info >= (3, 7):
+    _WrapperDescriptorType = types.WrapperDescriptorType
+else:
+    _WrapperDescriptorType = type(object.__init__)
 
 
 @verify.register(nodes.TypeInfo)
@@ -254,13 +271,32 @@ def verify_typeinfo(
         yield Error(object_path, "is not a type", stub, runtime, stub_desc=repr(stub))
         return
 
-    # Check everything already defined in the stub
+    try:
+        class SubClass(runtime):  # type: ignore
+            pass
+    except TypeError:
+        # Enum classes are implicitly @final
+        if not stub.is_final and not issubclass(runtime, enum.Enum):
+            yield Error(
+                object_path,
+                "cannot be subclassed at runtime, but isn't marked with @final in the stub",
+                stub,
+                runtime,
+                stub_desc=repr(stub),
+            )
+    except Exception:
+        # The class probably wants its subclasses to do something special.
+        # Examples: ctypes.Array, ctypes._SimpleCData
+        pass
+
+    # Check everything already defined on the stub class itself (i.e. not inherited)
     to_check = set(stub.names)
-    # There's a reasonable case to be made that we should always check all dunders, but it's
-    # currently quite noisy. We could turn this into a denylist instead of an allowlist.
+    # Check all public things on the runtime class
     to_check.update(
         # cast to workaround mypyc complaints
-        m for m in cast(Any, vars)(runtime) if not m.startswith("_") or m in SPECIAL_DUNDERS
+        m
+        for m in cast(Any, vars)(runtime)
+        if not is_probably_private(m) and m not in IGNORABLE_CLASS_DUNDERS
     )
 
     for entry in sorted(to_check):
@@ -269,11 +305,22 @@ def verify_typeinfo(
             mangled_entry = "_{}{}".format(stub.name, entry)
         stub_to_verify = next((t.names[entry].node for t in stub.mro if entry in t.names), MISSING)
         assert stub_to_verify is not None
-        yield from verify(
-            stub_to_verify,
-            getattr(runtime, mangled_entry, MISSING),
-            object_path + [entry],
-        )
+        try:
+            runtime_attr = getattr(runtime, mangled_entry, MISSING)
+        except Exception:
+            # Catch all exceptions in case the runtime raises an unexpected exception
+            # from __getattr__ or similar.
+            continue
+        # Do not error for an object missing from the stub
+        # If the runtime object is a types.WrapperDescriptorType object
+        # and has a non-special dunder name.
+        # The vast majority of these are false positives.
+        if not (
+            isinstance(stub_to_verify, Missing)
+            and isinstance(runtime_attr, _WrapperDescriptorType)
+            and is_dunder(mangled_entry, exclude_special=True)
+        ):
+            yield from verify(stub_to_verify, runtime_attr, object_path + [entry])
 
 
 def _verify_static_class_methods(
@@ -514,7 +561,7 @@ class Signature(Generic[T]):
             return max(index for _, index in all_args[arg_name])
 
         def get_type(arg_name: str) -> mypy.types.ProperType:
-            with mypy.state.strict_optional_set(True):
+            with mypy.state.state.strict_optional_set(True):
                 all_types = [
                     arg.variable.type or arg.type_annotation for arg, _ in all_args[arg_name]
                 ]
@@ -584,6 +631,7 @@ def _verify_signature(
         if (
             runtime_arg.kind != inspect.Parameter.POSITIONAL_ONLY
             and stub_arg.variable.name.startswith("__")
+            and not is_dunder(function_name, exclude_special=True)  # noisy for dunder methods
         ):
             yield (
                 'stub argument "{}" should be positional or keyword '
@@ -673,19 +721,39 @@ def verify_funcitem(
         yield Error(object_path, "is inconsistent, " + message, stub, runtime)
 
     signature = safe_inspect_signature(runtime)
+    runtime_is_coroutine = inspect.iscoroutinefunction(runtime)
+
+    if signature:
+        stub_sig = Signature.from_funcitem(stub)
+        runtime_sig = Signature.from_inspect_signature(signature)
+        runtime_sig_desc = f'{"async " if runtime_is_coroutine else ""}def {signature}'
+        stub_desc = f'def {stub_sig!r}'
+    else:
+        runtime_sig_desc, stub_desc = None, None
+
+    # Don't raise an error if the stub is a coroutine, but the runtime isn't.
+    # That results in false positives.
+    # See https://github.com/python/typeshed/issues/7344
+    if runtime_is_coroutine and not stub.is_coroutine:
+        yield Error(
+            object_path,
+            'is an "async def" function at runtime, but not in the stub',
+            stub,
+            runtime,
+            stub_desc=stub_desc,
+            runtime_desc=runtime_sig_desc
+        )
+
     if not signature:
         return
 
-    stub_sig = Signature.from_funcitem(stub)
-    runtime_sig = Signature.from_inspect_signature(signature)
-
     for message in _verify_signature(stub_sig, runtime_sig, function_name=stub.name):
         yield Error(
             object_path,
             "is inconsistent, " + message,
             stub,
             runtime,
-            runtime_desc="def " + str(signature),
+            runtime_desc=runtime_sig_desc,
         )
 
 
@@ -706,6 +774,18 @@ def verify_var(
             yield Error(object_path, "is not present at runtime", stub, runtime)
         return
 
+    if (
+        stub.is_initialized_in_class
+        and is_read_only_property(runtime)
+        and (stub.is_settable_property or not stub.is_property)
+    ):
+        yield Error(
+            object_path,
+            "is read-only at runtime but not in the stub",
+            stub,
+            runtime
+        )
+
     runtime_type = get_mypy_type_of_runtime_value(runtime)
     if (
         runtime_type is not None
@@ -738,7 +818,14 @@ def verify_overloadedfuncdef(
         return
 
     if stub.is_property:
-        # We get here in cases of overloads from property.setter
+        # Any property with a setter is represented as an OverloadedFuncDef
+        if is_read_only_property(runtime):
+            yield Error(
+                object_path,
+                "is read-only at runtime but not in the stub",
+                stub,
+                runtime
+            )
         return
 
     if not is_probably_a_function(runtime):
@@ -781,7 +868,7 @@ def verify_typevarexpr(
         yield None
 
 
-def _verify_property(stub: nodes.Decorator, runtime: Any) -> Iterator[str]:
+def _verify_readonly_property(stub: nodes.Decorator, runtime: Any) -> Iterator[str]:
     assert stub.func.is_property
     if isinstance(runtime, property):
         return
@@ -856,7 +943,7 @@ def verify_decorator(
         yield Error(object_path, "is not present at runtime", stub, runtime)
         return
     if stub.func.is_property:
-        for message in _verify_property(stub, runtime):
+        for message in _verify_readonly_property(stub, runtime):
             yield Error(object_path, message, stub, runtime)
         return
 
@@ -896,18 +983,76 @@ def verify_typealias(
     )
 
 
-SPECIAL_DUNDERS = ("__init__", "__new__", "__call__", "__init_subclass__", "__class_getitem__")
+# ====================
+# Helpers
+# ====================
 
 
-def is_dunder(name: str, exclude_special: bool = False) -> bool:
-    """Returns whether name is a dunder name.
+IGNORED_MODULE_DUNDERS = frozenset(
+    {
+        "__file__",
+        "__doc__",
+        "__name__",
+        "__builtins__",
+        "__package__",
+        "__cached__",
+        "__loader__",
+        "__spec__",
+        "__annotations__",
+        "__path__",  # mypy adds __path__ to packages, but C packages don't have it
+        "__getattr__",  # resulting behaviour might be typed explicitly
+        # TODO: remove the following from this list
+        "__author__",
+        "__version__",
+        "__copyright__",
+    }
+)
 
-    :param exclude_special: Whether to return False for a couple special dunder methods.
+IGNORABLE_CLASS_DUNDERS = frozenset(
+    {
+        # Special attributes
+        "__dict__",
+        "__text_signature__",
+        "__weakref__",
+        "__del__",  # Only ever called when an object is being deleted, who cares?
+        "__hash__",
+        "__getattr__",  # resulting behaviour might be typed explicitly
+        "__setattr__",  # defining this on a class can cause worse type checking
+        # isinstance/issubclass hooks that type-checkers don't usually care about
+        "__instancecheck__",
+        "__subclasshook__",
+        "__subclasscheck__",
+        # Pickle methods
+        "__setstate__",
+        "__getstate__",
+        "__getnewargs__",
+        "__getinitargs__",
+        "__reduce_ex__",
+        "__reduce__",
+        # ctypes weirdness
+        "__ctype_be__",
+        "__ctype_le__",
+        "__ctypes_from_outparam__",
+        # mypy limitations
+        "__abstractmethods__",  # Classes with metaclass=ABCMeta inherit this attribute
+        "__new_member__",  # If an enum defines __new__, the method is renamed as __new_member__
+        "__dataclass_fields__",  # Generated by dataclasses
+        "__dataclass_params__",  # Generated by dataclasses
+        "__doc__",  # mypy's semanal for namedtuples assumes this is str, not Optional[str]
+        # typing implementation details, consider removing some of these:
+        "__parameters__",
+        "__origin__",
+        "__args__",
+        "__orig_bases__",
+        "__final__",
+        # Consider removing __slots__?
+        "__slots__",
+    }
+)
 
-    """
-    if exclude_special and name in SPECIAL_DUNDERS:
-        return False
-    return name.startswith("__") and name.endswith("__")
+
+def is_probably_private(name: str) -> bool:
+    return name.startswith("_") and not is_dunder(name)
 
 
 def is_probably_a_function(runtime: Any) -> bool:
@@ -918,6 +1063,10 @@ def is_probably_a_function(runtime: Any) -> bool:
     )
 
 
+def is_read_only_property(runtime: object) -> bool:
+    return isinstance(runtime, property) and runtime.fset is None
+
+
 def safe_inspect_signature(runtime: Any) -> Optional[inspect.Signature]:
     try:
         return inspect.signature(runtime)
@@ -942,7 +1091,16 @@ def is_subtype_helper(left: mypy.types.Type, right: mypy.types.Type) -> bool:
     ):
         # Pretend Literal[0, 1] is a subtype of bool to avoid unhelpful errors.
         return True
-    with mypy.state.strict_optional_set(True):
+
+    if (
+        isinstance(right, mypy.types.TypedDictType)
+        and isinstance(left, mypy.types.Instance)
+        and left.type.fullname == "builtins.dict"
+    ):
+        # Special case checks against TypedDicts
+        return True
+
+    with mypy.state.state.strict_optional_set(True):
         return mypy.subtypes.is_subtype(left, right)
 
 
@@ -1029,17 +1187,24 @@ def get_mypy_type_of_runtime_value(runtime: Any) -> Optional[mypy.types.Type]:
         return mypy.types.TupleType(items, fallback)
 
     fallback = mypy.types.Instance(type_info, [anytype() for _ in type_info.type_vars])
-    try:
-        # Literals are supposed to be only bool, int, str, bytes or enums, but this seems to work
-        # well (when not using mypyc, for which bytes and enums are also problematic).
-        return mypy.types.LiteralType(
-            value=runtime,
-            fallback=fallback,
-        )
-    except TypeError:
-        # Ask for forgiveness if we're using mypyc.
+
+    value: Union[bool, int, str]
+    if isinstance(runtime, bytes):
+        value = bytes_to_human_readable_repr(runtime)
+    elif isinstance(runtime, enum.Enum):
+        value = runtime.name
+    elif isinstance(runtime, (bool, int, str)):
+        value = runtime
+    else:
         return fallback
 
+    return mypy.types.LiteralType(value=value, fallback=fallback)
+
+
+# ====================
+# Build and entrypoint
+# ====================
+
 
 _all_stubs: Dict[str, nodes.MypyFile] = {}
 
@@ -1107,19 +1272,27 @@ def get_stub(module: str) -> Optional[nodes.MypyFile]:
     return _all_stubs.get(module)
 
 
-def get_typeshed_stdlib_modules(custom_typeshed_dir: Optional[str]) -> List[str]:
+def get_typeshed_stdlib_modules(
+    custom_typeshed_dir: Optional[str],
+    version_info: Optional[Tuple[int, int]] = None
+) -> List[str]:
     """Returns a list of stdlib modules in typeshed (for current Python version)."""
     stdlib_py_versions = mypy.modulefinder.load_stdlib_py_versions(custom_typeshed_dir)
-    packages = set()
+    if version_info is None:
+        version_info = sys.version_info[0:2]
     # Typeshed's minimum supported Python 3 is Python 3.6
     if sys.version_info < (3, 6):
         version_info = (3, 6)
-    else:
-        version_info = sys.version_info[0:2]
-    for module, versions in stdlib_py_versions.items():
-        minver, maxver = versions
-        if version_info >= minver and (maxver is None or version_info <= maxver):
-            packages.add(module)
+
+    def exists_in_version(module: str) -> bool:
+        assert version_info is not None
+        parts = module.split(".")
+        for i in range(len(parts), 0, -1):
+            current_module = ".".join(parts[:i])
+            if current_module in stdlib_py_versions:
+                minver, maxver = stdlib_py_versions[current_module]
+                return version_info >= minver and (maxver is None or version_info <= maxver)
+        return False
 
     if custom_typeshed_dir:
         typeshed_dir = Path(custom_typeshed_dir)
@@ -1132,7 +1305,7 @@ def get_typeshed_stdlib_modules(custom_typeshed_dir: Optional[str]) -> List[str]
         if path.stem == "__init__":
             path = path.parent
         module = ".".join(path.relative_to(stdlib_dir).parts[:-1] + (path.stem,))
-        if module.split(".")[0] in packages:
+        if exists_in_version(module):
             modules.append(module)
     return sorted(modules)
 
@@ -1169,7 +1342,8 @@ def test_stubs(args: argparse.Namespace, use_builtins_fixtures: bool = False) ->
     if args.check_typeshed:
         assert not args.modules, "Cannot pass both --check-typeshed and a list of modules"
         modules = get_typeshed_stdlib_modules(args.custom_typeshed_dir)
-        annoying_modules = {"antigravity", "this"}
+        # typeshed added a stub for __main__, but that causes stubtest to check itself
+        annoying_modules = {"antigravity", "this", "__main__"}
         modules = [m for m in modules if m not in annoying_modules]
 
     assert modules, "No modules to check"
