@@ -12,18 +12,18 @@ other modules refer to them.
 """
 
 from abc import abstractmethod
-from mypy.ordered_dict import OrderedDict
+from mypy.backports import OrderedDict
 from typing import Generic, TypeVar, cast, Any, List, Callable, Iterable, Optional, Set, Sequence
 from mypy_extensions import trait, mypyc_attr
 
 T = TypeVar('T')
 
 from mypy.types import (
-    Type, AnyType, CallableType, Overloaded, TupleType, TypeGuardType, TypedDictType, LiteralType,
-    RawExpressionType, Instance, NoneType, TypeType,
-    UnionType, TypeVarType, PartialType, DeletedType, UninhabitedType, TypeVarLikeDef,
+    Type, AnyType, CallableType, Overloaded, TupleType, TypedDictType, LiteralType,
+    Parameters, RawExpressionType, Instance, NoneType, TypeType,
+    UnionType, TypeVarType, PartialType, DeletedType, UninhabitedType, TypeVarLikeType,
     UnboundType, ErasedType, StarType, EllipsisType, TypeList, CallableArgument,
-    PlaceholderType, TypeAliasType, get_proper_type
+    PlaceholderType, TypeAliasType, ParamSpecType, UnpackType, get_proper_type
 )
 
 
@@ -63,6 +63,14 @@ class TypeVisitor(Generic[T]):
     def visit_type_var(self, t: TypeVarType) -> T:
         pass
 
+    @abstractmethod
+    def visit_param_spec(self, t: ParamSpecType) -> T:
+        pass
+
+    @abstractmethod
+    def visit_parameters(self, t: Parameters) -> T:
+        pass
+
     @abstractmethod
     def visit_instance(self, t: Instance) -> T:
         pass
@@ -104,7 +112,7 @@ class TypeVisitor(Generic[T]):
         pass
 
     @abstractmethod
-    def visit_type_guard_type(self, t: TypeGuardType) -> T:
+    def visit_unpack_type(self, t: UnpackType) -> T:
         pass
 
 
@@ -167,7 +175,7 @@ class TypeTranslator(TypeVisitor[Type]):
         return t
 
     def visit_instance(self, t: Instance) -> Type:
-        last_known_value = None  # type: Optional[LiteralType]
+        last_known_value: Optional[LiteralType] = None
         if t.last_known_value is not None:
             raw_last_known_value = t.last_known_value.accept(self)
             assert isinstance(raw_last_known_value, LiteralType)  # type: ignore
@@ -183,9 +191,18 @@ class TypeTranslator(TypeVisitor[Type]):
     def visit_type_var(self, t: TypeVarType) -> Type:
         return t
 
+    def visit_param_spec(self, t: ParamSpecType) -> Type:
+        return t
+
+    def visit_parameters(self, t: Parameters) -> Type:
+        return t.copy_modified(arg_types=self.translate_types(t.arg_types))
+
     def visit_partial_type(self, t: PartialType) -> Type:
         return t
 
+    def visit_unpack_type(self, t: UnpackType) -> Type:
+        return t.type.accept(self)
+
     def visit_callable_type(self, t: CallableType) -> Type:
         return t.copy_modified(arg_types=self.translate_types(t.arg_types),
                                ret_type=t.ret_type.accept(self),
@@ -224,16 +241,13 @@ class TypeTranslator(TypeVisitor[Type]):
     def translate_types(self, types: Iterable[Type]) -> List[Type]:
         return [t.accept(self) for t in types]
 
-    def visit_type_guard_type(self, t: TypeGuardType) -> Type:
-        return TypeGuardType(t.type_guard.accept(self))
-
     def translate_variables(self,
-                            variables: Sequence[TypeVarLikeDef]) -> Sequence[TypeVarLikeDef]:
+                            variables: Sequence[TypeVarLikeType]) -> Sequence[TypeVarLikeType]:
         return variables
 
     def visit_overloaded(self, t: Overloaded) -> Type:
-        items = []  # type: List[CallableType]
-        for item in t.items():
+        items: List[CallableType] = []
+        for item in t.items:
             new = item.accept(self)
             assert isinstance(new, CallableType)  # type: ignore
             items.append(new)
@@ -269,7 +283,7 @@ class TypeQuery(SyntheticTypeVisitor[T]):
         self.strategy = strategy
         # Keep track of the type aliases already visited. This is needed to avoid
         # infinite recursion on types like A = Union[int, List[A]].
-        self.seen_aliases = set()  # type: Set[TypeAliasType]
+        self.seen_aliases: Set[TypeAliasType] = set()
 
     def visit_unbound_type(self, t: UnboundType) -> T:
         return self.query_types(t.args)
@@ -298,6 +312,15 @@ class TypeQuery(SyntheticTypeVisitor[T]):
     def visit_type_var(self, t: TypeVarType) -> T:
         return self.query_types([t.upper_bound] + t.values)
 
+    def visit_param_spec(self, t: ParamSpecType) -> T:
+        return self.strategy([])
+
+    def visit_unpack_type(self, t: UnpackType) -> T:
+        return self.query_types([t.type])
+
+    def visit_parameters(self, t: Parameters) -> T:
+        return self.query_types(t.arg_types)
+
     def visit_partial_type(self, t: PartialType) -> T:
         return self.strategy([])
 
@@ -326,11 +349,8 @@ class TypeQuery(SyntheticTypeVisitor[T]):
     def visit_union_type(self, t: UnionType) -> T:
         return self.query_types(t.items)
 
-    def visit_type_guard_type(self, t: TypeGuardType) -> T:
-        return t.type_guard.accept(self)
-
     def visit_overloaded(self, t: Overloaded) -> T:
-        return self.query_types(t.items())
+        return self.query_types(t.items)
 
     def visit_type_type(self, t: TypeType) -> T:
         return t.item.accept(self)
@@ -350,7 +370,7 @@ class TypeQuery(SyntheticTypeVisitor[T]):
         Use the strategy to combine the results.
         Skip type aliases already visited types to avoid infinite recursion.
         """
-        res = []  # type: List[T]
+        res: List[T] = []
         for t in types:
             if isinstance(t, TypeAliasType):
                 # Avoid infinite recursion for recursive type aliases.
